package edu.uiuc.dsi;

import java.nio.ByteBuffer;
import java.util.Arrays;

public class Packet
{
	public static final byte PACKET_TYPE_CONNECTION_NAME = 0x01;
	public static final byte PACKET_TYPE_COMPONENT_DATA = 0x02;
	public static final byte PACKET_TYPE_USER_DATA = 0x03;
	
	private byte packetType = 0x00;
	private String targetComponentName = null;
	private byte[] data;
	private boolean isCompletePacket;
	

	public static Packet parsePacket(byte[] data, int dataLength)
	{
		Packet objPacket = new Packet();
		
		if (dataLength >= 1)
		{
			objPacket.setPacketType(data[0]);
			
			int packetDataLength;
			short componentNameLength;
			
			switch (objPacket.getPacketType())
			{
				case PACKET_TYPE_COMPONENT_DATA:
					objPacket.setCompletePacketStatus(false);
					
					if (dataLength >= 2)
					{
						componentNameLength = (short)data[1];
						
						if (dataLength >= 2 + componentNameLength)
						{
							objPacket.setTargetComponentName(new String(Arrays.copyOfRange(data, 2, componentNameLength + 2)));
							
							if (dataLength >= 2 + componentNameLength + 2)
							{
								packetDataLength = (data[2 + componentNameLength] << 8) + data[3 + componentNameLength];
								
								if (dataLength >= 2 + componentNameLength + 2 + packetDataLength)
								{
									objPacket.setData(Arrays.copyOfRange(data, 4 + componentNameLength, 4 + componentNameLength + packetDataLength));
									objPacket.setCompletePacketStatus(true);
								}
							}
						}
					}
					break;
					
				case PACKET_TYPE_CONNECTION_NAME:
				case PACKET_TYPE_USER_DATA:
					objPacket.setCompletePacketStatus(false);
					
					if (dataLength >= 3)
					{
						packetDataLength = (data[1] << 8) + data[2];
						
						if (dataLength >= 3 + packetDataLength)
						{
							objPacket.setData(Arrays.copyOfRange(data, 3, 3 + packetDataLength));
							objPacket.setCompletePacketStatus(true);
						}
					}
					
					break;
			}
		}

		return objPacket;
	}
	
	public Packet()
	{
		this.isCompletePacket = false;
	}
	
	public Packet(byte[] data)
	{
		this.data = data;
		this.isCompletePacket = true;
	}
	
	public Packet(byte[] data, byte packetType)
	{
		this.data = data;
		this.packetType = packetType;
		this.isCompletePacket = true;
	}
	
	
	public byte[] getData()
	{
		return data;
	}
	
	public void setData(byte[] data)
	{
		this.data = data; 
	}
	
	public boolean isComponentPacket()
	{
		return (packetType == PACKET_TYPE_COMPONENT_DATA);
	}
	
	public String getTargetComponentName()
	{
		return targetComponentName;
	}

	public void setTargetComponentName(String name)
	{
		targetComponentName = name;
		
		if (name == null)
			packetType = PACKET_TYPE_USER_DATA;
		else
			packetType = PACKET_TYPE_COMPONENT_DATA;
	}
	
	public byte getPacketType()
	{
		return packetType;
	}
	
	public void setPacketType(byte packetType)
	{
		this.packetType = packetType;
	}
	
	public byte[] getFullPacketBytes()
	{
		ByteBuffer objByteBuffer = ByteBuffer.allocate(this.getFullPacketLength());
		
		if (isComponentPacket())
		{
			objByteBuffer.put((byte)packetType);
			objByteBuffer.put((byte)targetComponentName.length());
			objByteBuffer.put(targetComponentName.getBytes());
		}
		else
			objByteBuffer.put((byte)packetType);
		
		objByteBuffer.put((byte)((data.length >> 8) & 0xff));
		objByteBuffer.put((byte)((data.length) & 0xff));
		objByteBuffer.put(data);
		
		return objByteBuffer.array();
	}
	
	public boolean isCompletePacket()
	{
		return isCompletePacket;
	}
	
	public void setCompletePacketStatus(boolean value)
	{
		isCompletePacket = value;
	}
	
	public int getFullPacketLength()
	{
		if (isCompletePacket())
		{
			// Initial ID
			int iLength = 1;
			
			// Component
			if (isComponentPacket())
				iLength += 1 + targetComponentName.length();
			
			// Data Length Field
			iLength += 2;
			
			// Data
			iLength += data.length;
			
			return iLength;
		}
		else
			return 0;
	}

}
