package edu.uiuc.dsi.mutex;

import java.nio.ByteBuffer;
import java.util.ArrayList;

import edu.uiuc.dsi.Packet;
import edu.uiuc.dsi.TCPSocket;
import edu.uiuc.dsi.clocks.LamportClock;

public class RicartAgrawala extends Base
{
	private static final byte MESSAGE_REQUEST = (byte)0x01;
	private static final byte MESSAGE_REPLY = (byte)0x02;

	private LamportClock objLamportClock = new LamportClock();
	private ArrayList<TCPSocket> objTCPSocketsUnreplied = new ArrayList<TCPSocket>();
	private ArrayList<TCPSocket> objTCPSocketsDeferredReplies = new ArrayList<TCPSocket>();
	private LamportQueueElement objSelfMutExRequest = null;
	
	public RicartAgrawala(String componentName)
	{
		super(componentName);
	}

	@Override
	protected void doBeginMutExTasks()
	{
		objLamportClock.tick();
	}

	@Override
	protected void doReleaseMutExTasks()
	{
		objLamportClock.tick();
		objSelfMutExRequest = null;
		
		ByteBuffer objByteBuffer = ByteBuffer.allocate(3);
		objByteBuffer.put(MESSAGE_REPLY);
		objByteBuffer.putShort(objLamportClock.getCurrentTime());

		for (TCPSocket objTCPSocket : objTCPSocketsDeferredReplies)
			write(objTCPSocket, objByteBuffer.array());
		objTCPSocketsDeferredReplies.clear();
	}

	@Override
	protected void doRequestMutExTasks()
	{
		objLamportClock.tick();
		objSelfMutExRequest = new LamportQueueElement(objLamportClock.getCurrentTime(), this.getSelfTCPSocket());
		
		// Mark all sites as needing to be heard from...
		objTCPSocketsUnreplied.add(this.getSelfTCPSocket());
		for (TCPSocket objTCPSocket : getActiveTCPSockets())
			objTCPSocketsUnreplied.add(objTCPSocket); 
		
		// Broadcast request to all connected sites...
		ByteBuffer objByteBuffer = ByteBuffer.allocate(3);
		objByteBuffer.put(MESSAGE_REQUEST);
		objByteBuffer.putShort(objLamportClock.getCurrentTime());
		broadcast(objByteBuffer.array());		
	}

	@Override
	public void processPacket(TCPSocket sender, Packet packet)
	{
		ByteBuffer objByteBuffer = ByteBuffer.wrap(packet.getData());
		byte packetType = objByteBuffer.get();
		short packetTimestamp = 0;
		
		if (packetType == MESSAGE_REQUEST)
		{
			// Update Clock...
			packetTimestamp = objByteBuffer.getShort();
			objLamportClock.tick(packetTimestamp);

			if (objSelfMutExRequest == null || objSelfMutExRequest.compareTo(new LamportQueueElement(packetTimestamp, sender)) <= 0)
			{
				ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(3);
				objResponseByteBuffer.put(MESSAGE_REPLY);
				objResponseByteBuffer.putShort(objLamportClock.getCurrentTime());
				write(sender, objResponseByteBuffer.array());
			}
			else
				objTCPSocketsDeferredReplies.add(sender);
		}
		else if (packetType == MESSAGE_REPLY)
		{
			// 1. Update local clock based on remote site...
			packetTimestamp = objByteBuffer.getShort();
			objLamportClock.tick(packetTimestamp);
			
			// 2. Remove the REPLIED site from the list of sites waiting for a REPLY from...
			objTCPSocketsUnreplied.remove(sender);
			
			// 3. Begin the critical section if the Ricart-Agawala condition is met...
			if (objTCPSocketsUnreplied.size() == 0)
			{
				try { beginMutEx(); }
				catch (MutExException e) { e.printStackTrace(); }
			}
		}
		
		this.getApplicationCore().writeStatusMessage(this, "Status Update:");
		
		this.getApplicationCore().writeStatusMessage(this, "  Current State     = " + this.getCurrentMutualExclusionState());
		this.getApplicationCore().writeStatusMessage(this, "  Packet Type       = " + packetType);
		this.getApplicationCore().writeStatusMessage(this, "  Local Clock       = " + objLamportClock.getCurrentTime());
		this.getApplicationCore().writeStatusMessage(this, "  Packet Clock      = " + packetTimestamp);
		this.getApplicationCore().writeStatusMessage(this, "  Sender            = " + sender.getSocketName());
		
		this.getApplicationCore().writeStatusMessage(this, "  Unreplied Objects = " + objTCPSocketsUnreplied.size());
		for (int i=0; i < objTCPSocketsUnreplied.size(); i++)
			this.getApplicationCore().writeStatusMessage(this, "    " + objTCPSocketsUnreplied.get(i).getSocketName());

		this.getApplicationCore().writeStatusMessage(this, "  Deferred Objects  = " + objTCPSocketsDeferredReplies.size());
		for (int i=0; i < objTCPSocketsDeferredReplies.size(); i++)
			this.getApplicationCore().writeStatusMessage(this, "    " + objTCPSocketsDeferredReplies.get(i).getSocketName());

		if (objSelfMutExRequest == null)
			this.getApplicationCore().writeStatusMessage(this, "  MyObject TS       = null");
		else
			this.getApplicationCore().writeStatusMessage(this, "  MyObject TS       = " + objSelfMutExRequest.getTimestamp());
		
	
		
	}
}